# -*- coding: utf-8 -*-
"""
Backend values patchs
=====================

Patchs are used to modify given settings items like expanding paths. Backends
inherit from ``SettingsPatcher`` to be able to use it in their ``clean()``
method.

Todo:
    * Should be named postprocess.SettingsPostProcessor because it contains
      patchs but will contain some validations soon.
    * Add validation that source_dir is not also set as a library dir (would
      cause some issues in resolving);
"""
import os

from boussole.conf import SETTINGS_MANIFEST


class SettingsPatcher(object):
    """
    Mixin object for all available patch methods to use in
    ``SETTINGS_MANIFEST``.
    """
    projectdir = ''

    def patch(self, settings):
        """
        Perform patchs on settings according to their rules in
        ``SETTINGS_MANIFEST``.

        Patchs are implemented in their own method that all have the same
        signature:

        * Get arguments: initial settings, item name and item value;
        * Return item value possibly patched;

        Args:
            settings (dict): Loaded settings.

        Returns:
            dict: Settings object patched.

        """
        for k, v in settings.items():
            # Search for patchs method for setting rule
            if k in SETTINGS_MANIFEST and \
               SETTINGS_MANIFEST[k].get('patchs', None) is not None:
                patchs = SETTINGS_MANIFEST[k]['patchs']

                # Apply each patch on value
                for method_name in patchs:
                    settings[k] = getattr(self, method_name)(settings, k, v)

        return settings

    def _patch_expand_path(self, settings, name, value):
        """
        Patch a path to expand home directory and make absolute path.

        Args:
            settings (dict): Initial settings unpatched.
            name (str): Setting name.
            value (str): Path to patch.

        Returns:
            str: Patched path to an absolute path.

        """
        if os.path.isabs(value):
            return os.path.normpath(value)

        # Expand home directory if any
        value = os.path.expanduser(value)

        # If the path is not yet an absolute directory, make it so from base
        # directory if not empty
        if not os.path.isabs(value) and self.projectdir:
            value = os.path.join(self.projectdir, value)

        return os.path.normpath(value)

    def _patch_expand_paths(self, settings, name, value):
        """
        Apply patch ``SettingsPatcher._patch_expand_path`` for each element
        in list.

        Args:
            settings (dict): Initial settings unpatched.
            name (str): Setting name.
            value (list): List of paths to patch.

        Returns:
            list: Patched path list to an absolute path.

        """
        return [self._patch_expand_path(settings, name, item)
                for item in value]



# -*- coding: utf-8 -*-
import os
import pytest

def test_inspector_reset(settings, inspector):
    """inspector.ScssInspector: Ensure inspection maps are well reset"""
    sourcepath = os.path.join(settings.sample_path, 'main_basic.scss')
    
    inspector.inspect(sourcepath)

    inspector.reset()
    
    assert inspector._CHILDREN_MAP  == {}
    assert inspector._PARENTS_MAP == {}
    assert list(inspector.children(sourcepath)) == []
    assert list(inspector.parents(sourcepath))  == []



"""
Unittests
=========

Tests are builded using **py.test**.

In data fixtures, not all settings and SASS stylesheets will compile with
libsass. Some of them contains some errors for test needs.

* ``settings.txt``: Is not an usable settings file, just a dummy file for base backend tests;
* ``settings.json``: Is not an usable settings file, its values are dummy pointing to path and files that does not exists;
* ``settings_polluted.json``: Is almost equivalent of ``settings_custom.json`` but polluted with invalid setting names;
* ``settings_custom.json``: Contain only valid settings, will compile;

For SASS stylesheets, they should all contains a comment when it contains errors.
"""


"""
Main entrance to commandline actions
"""
import click

from boussole.cli.version import version_command
from boussole.cli.compile import compile_command
from boussole.cli.watch import watch_command
from boussole.logs import BOUSSOLE_LOGGER_CONF, init_logger


# Help alias on '-h' argument
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', count=True)
@click.pass_context
def cli_frontend(ctx, verbose):
    """
    Boussole is a commandline interface to build SASS projects using libsass.

    Every project will need a config file containing all needed settings to
    build it.
    """
    # Limit verbosity from enabled levels
    if verbose > len(BOUSSOLE_LOGGER_CONF)-1:
        verbose = len(BOUSSOLE_LOGGER_CONF)-1
    # Verbosity is the inverse of logging levels
    levels = [item[0] for item in BOUSSOLE_LOGGER_CONF]
    levels.reverse()
    # Init the logger config
    root_logger = init_logger(levels[verbose])

    root_logger.info("Logging level fixed to '{}'".format(levels[verbose]))

    # Init the default context that will be passed to commands
    ctx.obj = {
        'verbosity': verbose,
        'logger': root_logger,
    }


# Attach commands methods to the main grouper
cli_frontend.add_command(version_command, name="version")
cli_frontend.add_command(compile_command, name="compile")
cli_frontend.add_command(watch_command, name="watch")



# -*- coding: utf-8 -*-
import os
import pytest

from boussole.exceptions import UnresolvablePath

def test_inspector_libs_empty(settings, inspector):
    """inspector.ScssInspector: Some lib components used but no given library"""
    sourcepath = os.path.join(settings.sample_path, 'main_using_libs.scss')
    
    sources = [
        sourcepath,
    ]
    
    # Some imports use libraries that was not given, so resolver raise 
    # an error
    with pytest.raises(UnresolvablePath):
        inspector.inspect(*sources)

def test_inspector_libs_children(settings, inspector):
    """inspector.ScssInspector: Children with some lib components used"""
    sourcepath = os.path.join(settings.sample_path, 'main_using_libs.scss')
    
    sources = [
        sourcepath,
    ]
    
    inspector.inspect(*sources, library_paths=settings.libraries_fixture_paths)

    children = list(inspector.children(sourcepath))
    assert children == [
        os.path.join(settings.sample_path, '_empty.scss'),
        os.path.join(settings.lib1_path, 'settings/_sample.scss'),
        os.path.join(settings.sample_path, '_vendor.scss'),
        os.path.join(settings.lib2_path, 'addons/_some_addon.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.lib1_path, 'components/_panels.scss'),
        os.path.join(settings.lib1_path, 'library_1_fullstack.scss'),
        os.path.join(settings.lib2_path, 'addons/_another_addon.scss'),
        os.path.join(settings.lib1_path, 'components/_sections.scss'),
        os.path.join(settings.sample_path, 'components/_webfont_icons.scss'),
        os.path.join(settings.sample_path, 'components/_webfont.scss'),
    ]



# -*- coding: utf-8 -*-
import os
import click

from boussole.conf.json_backend import SettingsBackendJson
from boussole.exceptions import SettingsBackendError
from boussole.finder import ScssFinder
from boussole.compiler import SassCompileHelper


@click.command()
@click.option('--config', default=None, metavar='PATH',
              help='Path to a Boussole config file',
              type=click.Path(exists=True))
@click.pass_context
def compile_command(context, config):
    """
    Compile Sass stylesheets to CSS
    """
    logger = context.obj['logger']
    logger.info("Building project")

    # Load settings file
    try:
        backend = SettingsBackendJson(basedir=os.getcwd())
        settings = backend.load(filepath=config)
    except SettingsBackendError as e:
        logger.error(e.message)
        raise click.Abort()

    logger.debug("* Project sources directory: {}".format(
                settings.SOURCES_PATH))
    logger.debug("* Project destination directory: {}".format(
                settings.TARGET_PATH))
    logger.debug("* Exclude patterns: {}".format(
                settings.EXCLUDES))

    # Find all sources with their destination path
    compilable_files = ScssFinder().mirror_sources(
        settings.SOURCES_PATH,
        targetdir=settings.TARGET_PATH,
        excludes=settings.EXCLUDES
    )

    compiler = SassCompileHelper()

    # Build all compilable stylesheets
    for src, dst in compilable_files:
        logger.debug("* Compile: {}".format(src))

        output_opts = {}
        success, message = compiler.safe_compile(settings, src, dst)

        if success:
            click.secho("* Compiled: {}".format(message), **output_opts)
        else:
            click.secho(message, fg='red')



# -*- coding: utf-8 -*-
import os
import json

import pytest

import click
from click.testing import CliRunner

from boussole.cli.console_script import cli_frontend


def test_cli_compile_fail_001(settings):
    """cli.compile: Testing basic compile fail on default config filename
       (does not exists)"""
    runner = CliRunner()

    # Temporary isolated current dir
    with runner.isolated_filesystem():
        test_cwd = os.getcwd()

        result = runner.invoke(cli_frontend, ['compile'])

        assert 'Aborted!' in result.output
        assert result.exit_code == 1


def test_cli_compile_fail_002(settings):
    """cli.compile: Testing basic compile fail on given path directory (not a
       filename) as config file"""
    runner = CliRunner()

    # Temporary isolated current dir
    with runner.isolated_filesystem():
        test_cwd = os.getcwd()

        config_arg = '--config={}'.format(test_cwd)

        result = runner.invoke(cli_frontend, ['compile', config_arg])

        assert 'Aborted!' in result.output
        assert result.exit_code == 1


def test_cli_compile_fail_003(settings):
    """cli.compile: Testing basic compile fail on invalid config file (invalid
       JSON)"""
    runner = CliRunner()

    # Temporary isolated current dir
    with runner.isolated_filesystem():
        test_cwd = os.getcwd()

        with open('settings.json', 'w') as f:
            f.write('Invalid settings file content')

        result = runner.invoke(cli_frontend, ['compile'])

        msg = """No JSON object could be decoded from file: {}/settings.json"""
        assert msg.format(test_cwd) in result.output
        assert 'Aborted!' in result.output
        assert result.exit_code == 1


def test_cli_compile_success_001(settings):
    """cli.compile: Testing compile success on basic config, a main SASS
       source and a partial source to ignore"""
    runner = CliRunner()

    # Temporary isolated current dir
    with runner.isolated_filesystem():
        test_cwd = os.getcwd()

        # Write a minimal config file
        with open('settings.json', 'w') as f:
            f.write(json.dumps({
                'SOURCES_PATH': '.',
                'TARGET_PATH': './css',
                'OUTPUT_STYLES': 'compact',
            }, indent=4))

        # Write a minimal main SASS source
        source = "\n".join((
            """/* Main sample */""",
            """#content{""",
            """    color: red;""",
            """    &.wide{""",
            """        margin: 50px 15px;""",
            """    }""",
            """}""",
        ))
        with open('main.scss', 'w') as f:
            f.write(source)

        # Write a partial SASS source
        source = "\n".join((
            """/* Partial source to ignore */""",
            """.toignore-partial{""",
            """    color: gold !important;""",
            """}""",
        ))
        with open('_toignore.scss', 'w') as f:
            f.write(source)

        # Invoke command action
        result = runner.invoke(cli_frontend, ['-vvvv', 'compile'])

        # Attempted compiled CSS
        css_attempted = "\n".join((
            """/* Main sample */""",
            """#content { color: red; }""",
            "",
            """#content.wide { margin: 50px 15px; }""",
            "",
        ))
        with open(os.path.join(test_cwd, "css", "main.css"), 'rb') as f:
            css_compiled = f.read()

        #print result.exit_code
        #print result.output
        #print type(result)
        #print os.listdir(os.path.join(test_cwd, "css"))

        # Command success signal exit
        assert result.exit_code == 0
        # Source is correctly compiled
        assert css_compiled == css_attempted
        # Partial source to ignore is ignored
        assert os.listdir(os.path.join(test_cwd, "css")) == ['main.css']



# -*- coding: utf-8 -*-
import os
import io
import pytest


def test_compiler_write_001(settings, compiler, temp_builds_dir):
    """compiler.SassCompileHelper.write_content: Just creating file with latin
       content"""
    filepath = temp_builds_dir.join('compiler_001')

    content = u"""Some sample latin text"""

    compiler.write_content(content, filepath.strpath)

    # Read file to compare
    with io.open(filepath.strpath, 'r', encoding='utf-8') as f:
        result = f.read()

    assert content == result


def test_compiler_write_002(settings, compiler, temp_builds_dir):
    """compiler.SassCompileHelper.write_content: Creating file with unicode
       content"""
    filepath = temp_builds_dir.join('compiler_002')

    content = u"""Some sample unicode text: フランス Furansu"""

    compiler.write_content(content, filepath.strpath)

    # Read file to compare
    with io.open(filepath.strpath, 'r', encoding='utf-8') as f:
        result = f.read()

    assert content == result


def test_compiler_write_003(settings, compiler, temp_builds_dir):
    """compiler.SassCompileHelper.write_content: Creating file into
       subdirectory"""
    filepath = temp_builds_dir.join('foo/bar/home.txt')

    content = u"""Some sample unicode text: フランス Furansu"""

    compiler.write_content(content, filepath.strpath)

    # Read file to compare
    with io.open(filepath.strpath, 'r', encoding='utf-8') as f:
        result = f.read()

    assert content == result



# -*- coding: utf-8 -*-
"""
.. _libsass-python: https://github.com/dahlia/libsass-python

SASS compile helper
===================

This is not a real compiler, just an helper wrapping common methods to compile
a SASS source from a project using `libsass-python`_.
"""
import click
import os
import io

import sass


class SassCompileHelper(object):
    """
    Sass compile helper mixin
    """
    def safe_compile(self, settings, sourcepath, destination):
        """
        Safe compile

        It won't raise compile error and instead return compile success state
        as a boolean with a message.

        It will create needed directory structure first if it contain some
        directories that does not allready exists.

        Todo:
            Maybe this can inherit from finder.

        Args:
            settings (boussole.conf.model.Settings): Project settings.
            sourcepath (str): Source file path to compile to CSS.
            destination (str): Destination path for compiled CSS.

        Returns:
            tuple: A tuple of (success state, message).

            * success state: is boolean weither the compile is a success
              or not;
            * message: Message accorded to success. If compile fails, the
              message will contains returned error from libsass, if success
              just the destination path.
        """
        try:
            content = sass.compile(
                filename=sourcepath,
                output_style=settings.OUTPUT_STYLES,
                source_comments=settings.SOURCE_COMMENTS,
                include_paths=settings.LIBRARY_PATHS,
            )
        except sass.CompileError as e:
            click.secho(e.message, fg='red')
            return False, e.message
        else:
            self.write_content(content, destination)
            return True, destination

    def write_content(self, content, destination):
        """
        Write given content to destination path.

        It will create needed directory structure first if it contain some
        directories that does not allready exists.

        Args:
            content (str): Content to write to target file.
            destination (str): Destination path for target file.

        Returns:
            str: Path where target file has been written.
        """
        directory = os.path.dirname(destination)

        if directory and not os.path.exists(directory):
            os.makedirs(directory)

        with io.open(destination, 'w', encoding='utf-8') as f:
            f.write(content)

        return destination



# -*- coding: utf-8 -*-
import pytest


def test_resolver_candidate_basic(settings, parser, resolver):
    """resolver.ImportPathsResolver: Underscore leading and candidate extensions"""
    assert resolver.candidate_paths("foo") == [
        "foo.scss",
        "_foo.scss",
        "foo.sass",
        "_foo.sass",
        "foo.css",
        "_foo.css",
    ]


def test_resolver_candidate_extension_uncandidate(settings, parser, resolver):
    """resolver.ImportPathsResolver: Uncandidate extension"""
    assert resolver.candidate_paths("foo.plop") == [
        "foo.plop.scss",
        "_foo.plop.scss",
        "foo.plop.sass",
        "_foo.plop.sass",
        "foo.plop.css",
        "_foo.plop.css",
    ]


def test_resolver_candidate_extension_ready(settings, parser, resolver):
    """resolver.ImportPathsResolver: Candidate extension allready in place"""
    assert resolver.candidate_paths("foo.scss") == [
        "foo.scss",
        "_foo.scss",
    ]


def test_resolver_candidate_complex_001(settings, parser, resolver):
    """resolver.ImportPathsResolver: Complex case 1 for candidates"""
    assert resolver.candidate_paths("components/addons/foo.plop.scss") == [
        "components/addons/foo.plop.scss",
        "components/addons/_foo.plop.scss",
    ]


def test_resolver_candidate_complex_002(settings, parser, resolver):
    """resolver.ImportPathsResolver: Complex case 2 for candidates"""
    assert resolver.candidate_paths("../components/../addons/foo.plop") == [
        "../components/../addons/foo.plop.scss",
        "../components/../addons/_foo.plop.scss",
        "../components/../addons/foo.plop.sass",
        "../components/../addons/_foo.plop.sass",
        "../components/../addons/foo.plop.css",
        "../components/../addons/_foo.plop.css",
    ]



# -*- coding: utf-8 -*-
import os
import pytest

from boussole.exceptions import FinderException


def test_finder_relative_from_paths_001(finder):
    results = finder.get_relative_from_paths("/home/foo/plop", [
        "/home/foo",
        "/home/bar",
        "/etc",
    ])

    assert results == "plop"


def test_finder_relative_from_paths_002(finder):
    results = finder.get_relative_from_paths("/etc/plop.plip", [
        "/home/foo",
        "/home/bar",
        "/etc",
    ])

    assert results == "plop.plip"


def test_finder_relative_from_paths_003(finder):
    results = finder.get_relative_from_paths("/home/foo/plop", [
        "/home",
        "/home/foo",
        "/etc",
    ])

    assert results == "plop"


def test_finder_relative_from_paths_004(finder):
    results = finder.get_relative_from_paths("/home/foo/plop", [
        "/home",
        "/home/foo",
        "/home/bar",
        "/etc/ping",
    ])

    assert results == "plop"


def test_finder_relative_from_paths_005(finder):
    results = finder.get_relative_from_paths("/home/foo/plop", [
        "/home",
        "/home/foo",
        "/home/bar/pika",
        "/etc/ping",
    ])

    assert results == "plop"


def test_finder_relative_from_paths_006(finder):
    results = finder.get_relative_from_paths("/home/foo/pika/plop", [
        "/home",
        "/home/foo",
        "/home/bar/pika",
        "/home/bar",
    ])

    assert results == "pika/plop"


def test_finder_relative_from_paths_007(finder):
    results = finder.get_relative_from_paths("/home/foo/pika/plop", [
        "/etc",
        "/home/foo/pika",
        "/home/bar/pika",
        "/home/bar",
    ])

    assert results == "plop"


def test_finder_relative_from_paths_008(finder):
    results = finder.get_relative_from_paths("/home/foo/pika/bim/bam/plop", [
        "/etc",
        "/home/foo/pika/bim/bam",
        "/home/foo/pika/bim/bom",
        "/home/bar/pika",
        "/home/bar",
    ])

    assert results == "plop"


def test_finder_relative_from_paths_009(finder):
    """Unable to find relative path raise an exception"""
    with pytest.raises(FinderException):
        results = finder.get_relative_from_paths("/home/foo/pika/bim/bam/plop", [
            "/etc",
            "/home/foo/pika/bim/bom",
            "/home/bar/pika",
            "/home/bar",
        ])



# -*- coding: utf-8 -*-
import os
import logging
import pytest

from boussole.exceptions import UnresolvablePath

from utils import (DummyBaseEvent, DummyMoveEvent, DummyBaseHandler,
                   UnitTestableLibraryEventHandler,
                   UnitTestableProjectEventHandler, start_env,
                   build_sample_structure)


def test_watcher_project_compilablefiles_001(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: Testing 'handler.compilable_files' return"""
    basedir = temp_builds_dir.join('watcher_001')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    # Manually call on_any_event since we directly access to
    # 'handler.compilable_files' bypassing the event hierarchy
    project_handler.on_any_event(object())

    assert project_handler.compilable_files == {
        bdir('sass/main.scss'): bdir('css/main.css'),
        bdir('sass/main_importing.scss'): bdir('css/main_importing.css'),
        bdir('sass/main_usinglib.scss'): bdir('css/main_usinglib.css'),
    }


def test_watcher_project_move_010(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: 'Move' event on main sample"""
    basedir = temp_builds_dir.join('watcher_010')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    project_handler.on_moved(DummyMoveEvent(bdir('sass/main.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['main.css', 'main_importing.css']


def test_watcher_project_move_011(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: 'Move' event from a source depending from
       main sample"""
    basedir = temp_builds_dir.join('watcher_011')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    project_handler.on_moved(DummyMoveEvent(bdir('sass/main_importing.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['main_importing.css']


def test_watcher_project_move_012(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: 'Move' event on included partial
       source"""
    basedir = temp_builds_dir.join('watcher_012')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    project_handler.on_moved(DummyMoveEvent(bdir('sass/_toinclude.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['main_usinglib.css', 'main.css', 'main_importing.css']


def test_watcher_project_modified_020(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: 'Modified' event on included partial
       source"""
    basedir = temp_builds_dir.join('watcher_020')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    project_handler.on_modified(DummyBaseEvent(bdir('sass/_toinclude.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['main_usinglib.css', 'main.css', 'main_importing.css']


def test_watcher_project_created_030(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: 'Created' event for a new main source"""
    basedir = temp_builds_dir.join('watcher_030')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    # Write a new main source
    source = "\n".join((
        """/* New main source */""",
        """#content{ padding: 10px; margin: 5px 0; }""",
    ))
    with open(basedir.join('sass/new_main.scss').strpath, 'w') as f:
        f.write(source)

    project_handler.on_created(DummyBaseEvent(bdir('sass/new_main.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['new_main.css']


def test_watcher_project_deleted_040(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: 'Deleted' event for a main source"""
    basedir = temp_builds_dir.join('watcher_040')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    os.remove(bdir('sass/main_importing.scss'))

    project_handler.on_deleted(DummyBaseEvent(bdir('sass/main_importing.scss')))
    assert os.listdir(basedir.join("css").strpath) == []


def test_watcher_project_deleted_041(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: 'Deleted' event for a partial source
       included by other files"""
    basedir = temp_builds_dir.join('watcher_041')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    os.remove(bdir('sass/_toinclude.scss'))

    with pytest.raises(UnresolvablePath):
        project_handler.on_deleted(DummyBaseEvent(bdir('sass/_toinclude.scss')))


def test_watcher_project_whole_050(settings, temp_builds_dir):
    """watcher.SassProjectEventHandler: Routine using some events on various
       sources"""
    basedir = temp_builds_dir.join('watcher_050')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableProjectEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    # Modify a partial source
    project_handler.on_modified(DummyBaseEvent(bdir('sass/_toinclude.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['main_usinglib.css', 'main.css', 'main_importing.css']

    # Create a new main source
    source = "\n".join((
        """/* New main source */""",
        """#content{ padding: 10px; margin: 5px 0; }""",
    ))
    with open(basedir.join('sass/new_main.scss').strpath, 'w') as f:
        f.write(source)

    project_handler.on_created(DummyBaseEvent(bdir('sass/new_main.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['new_main.css', 'main_usinglib.css', 'main.css', 'main_importing.css']

    # Delete a main source
    os.remove(bdir('css/main_importing.css'))
    os.remove(bdir('sass/main_importing.scss'))

    project_handler.on_deleted(DummyBaseEvent(bdir('sass/main_importing.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['new_main.css', 'main_usinglib.css', 'main.css']

    # Simulate moved source
    project_handler.on_moved(DummyMoveEvent(bdir('sass/_toinclude.scss')))
    assert os.listdir(basedir.join("css").strpath) == ['new_main.css', 'main_usinglib.css', 'main.css']


def test_watcher_library_modified_101(settings, temp_builds_dir):
    """watcher.SassLibraryEventHandler: 'Modified' event on a library
       component"""
    basedir = temp_builds_dir.join('watcher_101')

    bdir, logger, inspector, settings_object, watcher_opts = start_env(basedir)

    build_sample_structure(settings_object, basedir)

    # Init handler
    project_handler = UnitTestableLibraryEventHandler(
        settings_object,
        logger,
        inspector,
        **watcher_opts
    )

    project_handler.on_modified(DummyBaseEvent(bdir('lib/components/_buttons.scss')))

    project_handler.on_modified(DummyBaseEvent(bdir('lib/libmain.scss')))

    # Almost dummy validation because on wrong behavior (with
    # UnitTestableProjectEventHandler instead of
    # UnitTestableLibraryEventHandler) CSS files are writed to "../lib"
    # path, that is under the "css" dir.
    assert os.listdir(basedir.join("css").strpath) == ['main_usinglib.css']



# -*- coding: utf-8 -*-
import os
import pytest


def test_finder_compilable_relative_nonrecursive_001(settings, finder):
    """finder.ScssFinder: Find non recursively all compilable sources
       from sample project in relative mode"""
    assert finder.compilable_sources(settings.sample_path, recursive=False) == [
        'main_syntax.scss',
        'main_error.scss',
        'main_circular_5.scss',
        'main_depth_import-2.scss',
        'main_twins_2.scss',
        'main_twins_1.scss',
        'main_circular_0.scss',
        'main_circular_4.scss',
        'main_circular_3.scss',
        'main_twins_3.scss',
        'main_using_libs.scss',
        'main_depth_import-3.scss',
        'main_circular_1.scss',
        'main_depth_import-1.scss',
        'main_commented.scss',
        'main_with_subimports.scss',
        'main_circular_2.scss',
        'main_circular_bridge.scss',
        'main_basic.scss',
    ]


def test_finder_compilable_absolute_nonrecursive_002(settings, finder):
    """finder.ScssFinder: Find non recursively all compilable sources
       from sample project in absolute mode"""
    assert finder.compilable_sources(settings.sample_path, absolute=True, recursive=False) == [
        os.path.join(settings.sample_path, 'main_syntax.scss'),
        os.path.join(settings.sample_path, 'main_error.scss'),
        os.path.join(settings.sample_path, 'main_circular_5.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-2.scss'),
        os.path.join(settings.sample_path, 'main_twins_2.scss'),
        os.path.join(settings.sample_path, 'main_twins_1.scss'),
        os.path.join(settings.sample_path, 'main_circular_0.scss'),
        os.path.join(settings.sample_path, 'main_circular_4.scss'),
        os.path.join(settings.sample_path, 'main_circular_3.scss'),
        os.path.join(settings.sample_path, 'main_twins_3.scss'),
        os.path.join(settings.sample_path, 'main_using_libs.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-3.scss'),
        os.path.join(settings.sample_path, 'main_circular_1.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-1.scss'),
        os.path.join(settings.sample_path, 'main_commented.scss'),
        os.path.join(settings.sample_path, 'main_with_subimports.scss'),
        os.path.join(settings.sample_path, 'main_circular_2.scss'),
        os.path.join(settings.sample_path, 'main_circular_bridge.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
    ]


def test_finder_compilable_relative_recursive_003(settings, finder):
    """finder.ScssFinder: Find recursively all compilable sources from
       sample project in relative mode"""
    assert finder.compilable_sources(settings.sample_path) == [
        'main_syntax.scss',
        'main_error.scss',
        'main_circular_5.scss',
        'main_depth_import-2.scss',
        'main_twins_2.scss',
        'main_twins_1.scss',
        'main_circular_0.scss',
        'main_circular_4.scss',
        'main_circular_3.scss',
        'main_twins_3.scss',
        'main_using_libs.scss',
        'main_depth_import-3.scss',
        'main_circular_1.scss',
        'main_depth_import-1.scss',
        'main_commented.scss',
        'main_with_subimports.scss',
        'main_circular_2.scss',
        'main_circular_bridge.scss',
        'main_basic.scss',
        'components/twin_3.scss',
        'components/twin_2.scss',
    ]


def test_finder_compilable_absolute_recursive_004(settings, finder):
    """finder.ScssFinder: Find recursively all compilable sources from
       sample project in absolute mode"""
    assert finder.compilable_sources(settings.sample_path, absolute=True) == [
        os.path.join(settings.sample_path, 'main_syntax.scss'),
        os.path.join(settings.sample_path, 'main_error.scss'),
        os.path.join(settings.sample_path, 'main_circular_5.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-2.scss'),
        os.path.join(settings.sample_path, 'main_twins_2.scss'),
        os.path.join(settings.sample_path, 'main_twins_1.scss'),
        os.path.join(settings.sample_path, 'main_circular_0.scss'),
        os.path.join(settings.sample_path, 'main_circular_4.scss'),
        os.path.join(settings.sample_path, 'main_circular_3.scss'),
        os.path.join(settings.sample_path, 'main_twins_3.scss'),
        os.path.join(settings.sample_path, 'main_using_libs.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-3.scss'),
        os.path.join(settings.sample_path, 'main_circular_1.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-1.scss'),
        os.path.join(settings.sample_path, 'main_commented.scss'),
        os.path.join(settings.sample_path, 'main_with_subimports.scss'),
        os.path.join(settings.sample_path, 'main_circular_2.scss'),
        os.path.join(settings.sample_path, 'main_circular_bridge.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.sample_path, 'components/twin_3.scss'),
        os.path.join(settings.sample_path, 'components/twin_2.scss'),
    ]


def test_finder_compilable_relative_subdir_005(settings, finder):
    """finder.ScssFinder: Find recursively all compilable sources from
       sample project subdirectory in relative mode"""
    assert finder.compilable_sources(os.path.join(settings.sample_path, 'components')) == [
        'twin_3.scss',
        'twin_2.scss',
    ]


def test_finder_compilable_absolute_subdir_006(settings, finder):
    """finder.ScssFinder: Find recursively all compilable sources from
       sample project subdirectory in absolute mode"""
    assert finder.compilable_sources(os.path.join(settings.sample_path, 'components'), absolute=True) == [
        os.path.join(settings.sample_path, 'components/twin_3.scss'),
        os.path.join(settings.sample_path, 'components/twin_2.scss'),
    ]


def test_finder_compilable_relative_exclude_007(settings, finder):
    """finder.ScssFinder: Find recursively all compilable sources with
       excluding patterns on relative mode"""
    excludes = [
        'main_error.scss',
        'main_circular_5.scss',
        'main_circular_0.scss',
        'main_circular_4.scss',
        'main_circular_3.scss',
        'main_circular_1.scss',
        'main_circular_2.scss',
        'main_circular_bridge.scss',
        '*/twin_3.scss',
    ]

    assert finder.compilable_sources(settings.sample_path, excludes=excludes) == [
        'main_syntax.scss',
        'main_depth_import-2.scss',
        'main_twins_2.scss',
        'main_twins_1.scss',
        'main_twins_3.scss',
        'main_using_libs.scss',
        'main_depth_import-3.scss',
        'main_depth_import-1.scss',
        'main_commented.scss',
        'main_with_subimports.scss',
        'main_basic.scss',
        'components/twin_2.scss',
    ]


def test_finder_compilable_absolute_exclude_008(settings, finder):
    """finder.ScssFinder: Find recursively all compilable sources with
       excluding patterns on absolute mode"""
    excludes = [
        'main_error.scss',
        'main_circular_5.scss',
        'main_circular_0.scss',
        'main_circular_4.scss',
        'main_circular_3.scss',
        'main_circular_1.scss',
        'main_circular_2.scss',
        'main_circular_bridge.scss',
        '*/twin_3.scss',
    ]

    assert finder.compilable_sources(settings.sample_path, absolute=True, excludes=excludes) == [
        os.path.join(settings.sample_path, 'main_syntax.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-2.scss'),
        os.path.join(settings.sample_path, 'main_twins_2.scss'),
        os.path.join(settings.sample_path, 'main_twins_1.scss'),
        os.path.join(settings.sample_path, 'main_twins_3.scss'),
        os.path.join(settings.sample_path, 'main_using_libs.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-3.scss'),
        os.path.join(settings.sample_path, 'main_depth_import-1.scss'),
        os.path.join(settings.sample_path, 'main_commented.scss'),
        os.path.join(settings.sample_path, 'main_with_subimports.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.sample_path, 'components/twin_2.scss'),
    ]



# -*- coding: utf-8 -*-
import pytest


def test_parser_flatten_rules_001(settings, parser):
    """parser.ScssImportsParser: flatten_rules case 1"""
    rules = parser.flatten_rules([
        ('', '"foo"'),
    ])
    assert rules == ['foo']


def test_parser_flatten_rules_002(settings, parser):
    """parser.ScssImportsParser: flatten_rules case 2"""
    rules = parser.flatten_rules([
        ('', "'bar'"),
    ])
    assert rules == ['bar']


def test_parser_flatten_rules_003(settings, parser):
    """parser.ScssImportsParser: flatten_rules case 3"""
    rules = parser.flatten_rules([
        ('', "'bar'"),
        ('url', '"wrong"'),
        ('', '"cool", "plop"'),
        ('', '"wrong.css"'),
        ('', '"https://wrong"'),
    ])
    assert rules == [
        'bar',
        'cool', 'plop',
    ]



# -*- coding: utf-8 -*-
import os
import copy
import pytest

from boussole.conf.model import Settings

def test_conf_settings_update_001_basic(settings, sample_project_settings):
    """conf.Settings: Ensure dummy update is correct"""
    settings_object = Settings()

    settings_object.update(sample_project_settings)

    assert settings_object._settings == sample_project_settings


def test_conf_settings_update_002_polluted(settings, sample_project_settings):
    """conf.Settings: Ensure update filtering is done"""
    settings_object = Settings(initial=sample_project_settings)

    # Add some wrong settings
    settings_object.update({
        'PLOP': True,
        'Foo': 42,
        'BAR': [1, 5, 10]
    })

    assert settings_object._settings == sample_project_settings


def test_conf_settings_update_003_mixed(settings, sample_project_settings):
    """conf.Settings: Ensure update filtering is correct on both enabled and
       not enabled setting names"""
    settings_object = Settings(initial=sample_project_settings)

    # Add one available setting and some other wrong settings
    settings_object.update({
        'OUTPUT_STYLES': 'compressed',
        'PLOP': True,
        'Foo': 42,
        'BAR': [1, 5, 10]
    })

    # New settings reference with modified setting
    mixed = copy.deepcopy(sample_project_settings)
    mixed.update({
        'OUTPUT_STYLES': 'compressed',
    })

    assert settings_object._settings == mixed



# -*- coding: utf-8 -*-
import os
import copy
import pytest

from boussole.conf.json_backend import SettingsBackendJson


def test_conf_backend_json_load_001_basic(settings, sample_project_settings):
    """conf.json_backend.SettingsBackendJson: Load basic JSON settings file"""
    backend = SettingsBackendJson(basedir=settings.fixtures_path)

    settings_object = backend.load()

    assert settings_object._settings == sample_project_settings

    assert settings_object.TARGET_PATH == sample_project_settings['TARGET_PATH']
    assert settings_object.SOURCES_PATH == sample_project_settings['SOURCES_PATH']
    assert settings_object.LIBRARY_PATHS == sample_project_settings['LIBRARY_PATHS']
    assert settings_object.OUTPUT_STYLES == sample_project_settings['OUTPUT_STYLES']


def test_conf_backend_json_load_002_poluted(settings, sample_project_settings):
    """conf.json_backend.SettingsBackendJson: Load polluted JSON settings
       file"""
    backend = SettingsBackendJson(basedir=settings.fixtures_path)

    settings_object = backend.load(filepath="settings_polluted.json")

    assert settings_object._settings == sample_project_settings

    assert settings_object.TARGET_PATH == sample_project_settings['TARGET_PATH']
    assert settings_object.SOURCES_PATH == sample_project_settings['SOURCES_PATH']
    assert settings_object.LIBRARY_PATHS == sample_project_settings['LIBRARY_PATHS']
    assert settings_object.OUTPUT_STYLES == sample_project_settings['OUTPUT_STYLES']

    # Wrong settings that does not exist and should have been removed
    assert getattr(settings_object, 'FOO', None) == None
    assert getattr(settings_object, 'BAR', None) == None


def test_conf_backend_json_load_003_custom(settings, custom_project_settings):
    """conf.json_backend.SettingsBackendJson: Load custom JSON settings file"""
    backend = SettingsBackendJson(basedir=settings.fixtures_path)

    settings_object = backend.load(filepath="settings_custom.json")

    assert settings_object._settings == custom_project_settings

    assert settings_object.TARGET_PATH == custom_project_settings['TARGET_PATH']
    assert settings_object.SOURCES_PATH == custom_project_settings['SOURCES_PATH']
    assert settings_object.LIBRARY_PATHS == custom_project_settings['LIBRARY_PATHS']
    assert settings_object.OUTPUT_STYLES == custom_project_settings['OUTPUT_STYLES']

    # Wrong settings that does not exist and should have been removed
    assert getattr(settings_object, 'FOO', None) == None
    assert getattr(settings_object, 'BAR', None) == None


def test_conf_backend_json_load_004_custom(settings, custom_project_settings):
    """conf.json_backend.SettingsBackendJson: Load custom JSON settings file
       with basedir and relative filepath"""
    backend = SettingsBackendJson(basedir=settings.tests_path)

    filepath = os.path.join(settings.fixtures_dir, "settings_custom.json")

    settings_object = backend.load(filepath=filepath)

    assert settings_object._settings == custom_project_settings

    assert settings_object.TARGET_PATH == custom_project_settings['TARGET_PATH']
    assert settings_object.SOURCES_PATH == custom_project_settings['SOURCES_PATH']
    assert settings_object.LIBRARY_PATHS == custom_project_settings['LIBRARY_PATHS']
    assert settings_object.OUTPUT_STYLES == custom_project_settings['OUTPUT_STYLES']

    # Wrong settings that does not exist and should have been removed
    assert getattr(settings_object, 'FOO', None) == None
    assert getattr(settings_object, 'BAR', None) == None



# -*- coding: utf-8 -*-
import os
import pytest

from boussole.conf.base_backend import SettingsBackendBase
from boussole.conf.json_backend import SettingsBackendJson


def test_conf_backend_base_parsepath_001_empty_nobasedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: no path given and
       with empty basedir"""
    backend = SettingsBackendJson()

    result = backend.parse_filepath()

    assert result == (
        '.',
        "settings.json"
    )


def test_conf_backend_base_parsepath_002_empty_basedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: no path given and
       with a basedir"""
    backend = SettingsBackendJson(basedir="/home/bart/www")

    result = backend.parse_filepath()

    assert result == (
        "/home/bart/www",
        "settings.json"
    )


def test_conf_backend_base_parsepath_003_filename_nobasedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: filename and empty
       basedir"""
    backend = SettingsBackendJson()

    result = backend.parse_filepath(filepath="settings_custom.json")

    assert result == (
        '.',
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_004_filename_basedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: filename and filled
       basedir"""
    backend = SettingsBackendJson(basedir="/home/bart/www")

    result = backend.parse_filepath(filepath="settings_custom.json")

    assert result == (
        '/home/bart/www',
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_005_relative_nobasedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: relative filepath and empty
       basedir"""
    backend = SettingsBackendJson()

    result = backend.parse_filepath(filepath="foo/settings_custom.json")

    assert result == (
        'foo',
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_006_relative_basedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: relative filepath and filled
       basedir"""
    backend = SettingsBackendJson(basedir="/home/bart/www")

    result = backend.parse_filepath(filepath="foo/settings_custom.json")

    assert result == (
        "/home/bart/www/foo",
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_007_absolute_nobasedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: absolute filepath and empty
       basedir"""
    backend = SettingsBackendJson()

    result = backend.parse_filepath(filepath="/home/bart/www/settings_custom.json")

    assert result == (
        "/home/bart/www",
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_008_absolute_basedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: absolute filepath and filled
       basedir to ensure basedir is ignored with absolute filepath"""
    backend = SettingsBackendJson(basedir="/home/no/pasaran")

    result = backend.parse_filepath(filepath="/home/bart/www/settings_custom.json")

    assert result == (
        "/home/bart/www",
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_010_normalize_01_basedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: filename and filled basedir,
       need normalize"""
    backend = SettingsBackendJson(basedir="/home/bart/www")

    result = backend.parse_filepath(filepath="./settings_custom.json")

    assert result == (
        '/home/bart/www',
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_009_normalize_02_basedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: filename and filled basedir,
       need normalize"""
    backend = SettingsBackendJson(basedir="/home/bart/www")

    result = backend.parse_filepath(filepath="../settings_custom.json")

    assert result == (
        '/home/bart',
        "settings_custom.json",
    )


def test_conf_backend_base_parsepath_011_normalize_nobasedir(settings, sample_project_settings):
    """conf.SettingsBackendJson.parse_filepath: filename and empty basedir,
       normalize can't do anything"""
    backend = SettingsBackendJson()

    result = backend.parse_filepath(filepath="../settings_custom.json")

    assert result == (
        '..',
        "settings_custom.json",
    )



# -*- coding: utf-8 -*-
"""
Inspector
=========

Inspector is in charge to inspect a project about SASS stylesheets to search
for their dependencies.
"""
from collections import defaultdict

from boussole.exceptions import CircularImport
from boussole.parser import ScssImportsParser
from boussole.resolver import ImportPathsResolver


class ScssInspector(ImportPathsResolver, ScssImportsParser):
    """
    Project inspector for SCSS sources

    Inspector is stateful, meaning you will need to invoke ``reset()`` then
    ``inspect()`` each time a project change, else the parents and children
    maps will be eventually incorrects.

    ``__init__`` method use ``reset`` method to initialize some internal
    buffers.

    Attributes:
        _CHILDREN_MAP: Dictionnary of finded direct children for each
            inspected sources.
        _PARENTS_MAP: Dictionnary of finded direct parents for each inspected
            sources.
    """

    def __init__(self, *args, **kwargs):
        self.reset()

    def reset(self):
        """
        Reset internal buffers ``_CHILDREN_MAP`` and ``_PARENTS_MAP``.
        """
        self._CHILDREN_MAP = {}
        self._PARENTS_MAP = defaultdict(set)

    def look_source(self, sourcepath, library_paths=None):
        """
        Open a SCSS file (sourcepath) and find all involved file through
        imports.

        This will fill internal buffers ``_CHILDREN_MAP`` and ``_PARENTS_MAP``.

        Args:
            sourcepath (str): Source file path to start searching for imports.

        Keyword Arguments:
            library_paths (list): List of directory paths for libraries to
                resolve paths if resolving fails on the base source path.
                Default to None.
        """
        # Don't inspect again source that has allready be inspected as a
        # children of a previous source
        if sourcepath not in self._CHILDREN_MAP:
            with open(sourcepath) as fp:
                finded_paths = self.parse(fp.read())

            children = self.resolve(sourcepath, finded_paths,
                                    library_paths=library_paths)

            # Those files that are imported by the sourcepath
            self._CHILDREN_MAP[sourcepath] = children

            # Those files that import the sourcepath
            for p in children:
                self._PARENTS_MAP[p].add(sourcepath)

            # Start recursive finding through each resolved path that has not
            # been collected yet
            for path in children:
                if path not in self._CHILDREN_MAP:
                    self.look_source(path, library_paths=library_paths)

        return

    def inspect(self, *args, **kwargs):
        """
        Recursively inspect all given SCSS files to find imported dependencies.

        This does not return anything. Just fill internal buffers about
        inspected files.

        Note:
            This will ignore orphan files (files that are not imported from
            any of given SCSS files).

        Args:
            *args: One or multiple arguments, each one for a source file path
                to inspect.

        Keyword Arguments:
            library_paths (list): List of directory paths for libraries to
                resolve paths if resolving fails on the base source path.
                Default to None.
        """
        library_paths = kwargs.get('library_paths', None)

        for sourcepath in args:
            self.look_source(sourcepath, library_paths=library_paths)

    def _get_recursive_dependancies(self, dependencies_map, sourcepath,
                                    recursive=True):
        """
        Return all dependencies of a source, recursively searching through its
        dependencies.

        This is a common method used by ``children`` and ``parents`` methods.

        Args:
            dependencies_map (dict): Internal buffer (internal buffers
                ``_CHILDREN_MAP`` or ``_PARENTS_MAP``) to use for searching.
            sourcepath (str): Source file path to start searching for
                dependencies.

        Keyword Arguments:
            recursive (bool): Switch to enable recursive finding (if True).
                Default to True.

        Raises:
            CircularImport: If circular error is detected from a source.

        Returns:
            set: List of dependencies paths.
        """
        # Direct dependencies
        collected = set([])
        collected.update(dependencies_map.get(sourcepath, []))

        # Sequence of 'dependencies_map' items to explore
        sequence = collected.copy()
        # Exploration list
        walkthrough = []

        # Recursive search starting from direct dependencies
        if recursive:
            while True:
                if not sequence:
                    break
                item = sequence.pop()

                # Add current source to the explorated source list
                walkthrough.append(item)

                # Current item children
                current_item_dependancies = dependencies_map.get(item, [])

                for dependency in current_item_dependancies:
                    # Allready visited item, ignore and continue to the new
                    # item
                    if dependency in walkthrough:
                        continue
                    # Unvisited item yet, add its children to dependencies and
                    # item to explore
                    else:
                        collected.add(dependency)
                        sequence.add(dependency)

                # Sourcepath has allready been visited but present itself
                # again, assume it's a circular import
                if sourcepath in walkthrough:
                    msg = "A circular import has occured by '{}'"
                    raise CircularImport(msg.format(current_item_dependancies))

                # No more item to explore, break loop
                if not sequence:
                    break

        return collected

    def children(self, sourcepath, recursive=True):
        """
        Recursively find all children that are imported from the given source
        path.

        Args:
            sourcepath (str): Source file path to search for.

        Keyword Arguments:
            recursive (bool): Switch to enabled recursive finding (if True).
                Default to True.

        Returns:
            set: List of finded parents path.
        """
        return self._get_recursive_dependancies(
            self._CHILDREN_MAP,
            sourcepath,
            recursive=True
        )

    def parents(self, sourcepath, recursive=True):
        """
        Recursively find all parents that import the given source path.

        Args:
            sourcepath (str): Source file path to search for.

        Keyword Arguments:
            recursive (bool): Switch to enabled recursive finding (if True).
                Default to True.

        Returns:
            set: List of finded parents path.
        """
        return self._get_recursive_dependancies(
            self._PARENTS_MAP,
            sourcepath,
            recursive=True
        )



# -*- coding: utf-8 -*-
"""
Exceptions
==========

Specific exceptions that Boussole code can raise.
"""


class BoussoleBaseException(Exception):
    """
    Base for Boussole exceptions.
    """
    pass


class InvalidImportRule(BoussoleBaseException):
    """
    Exception to be raised when the parser encounts an invalid import rule.
    """
    pass


class FinderException(BoussoleBaseException):
    """
    Exception to be raised when error occurs with finder usage.
    """
    pass


class UnresolvablePath(BoussoleBaseException):
    """
    Exception to be raised when the resolver can not resolve a given path.
    """
    pass


class UnclearResolution(BoussoleBaseException):
    """
    Exception to be raised when the resolver encounts multiple existing
    candidates for a path.
    """
    pass


class CircularImport(BoussoleBaseException):
    """
    Exception to be raised when inspector detect a circular import from
    sources.
    """
    pass


class SettingsBackendError(BoussoleBaseException):
    """
    Exception to be raised when config loading has failed from a backend.
    """
    pass


class SettingsInvalidError(BoussoleBaseException):
    """
    Exception to be raised when a settings is detected as invalid.
    """
    pass



# -*- coding: utf-8 -*-
import time
import os
import click

from watchdog.observers import Observer

from boussole.conf.json_backend import SettingsBackendJson
from boussole.exceptions import SettingsBackendError
from boussole.inspector import ScssInspector
from boussole.watcher import (WatchdogLibraryEventHandler,
                              WatchdogProjectEventHandler)


@click.command()
@click.option('--config', default=None, metavar='PATH',
              help='Path to a Boussole config file',
              type=click.Path(exists=True))
@click.pass_context
def watch_command(context, config):
    """
    Watch for change on your project Sass stylesheets then compile them to CSS.
    """
    logger = context.obj['logger']
    logger.info("Watching project")

    # Load settings file
    try:
        backend = SettingsBackendJson(basedir=os.getcwd())
        settings = backend.load(filepath=config)
    except SettingsBackendError as e:
        logger.error(e.message)
        raise click.Abort()

    logger.debug("* Project sources directory: {}".format(
                settings.SOURCES_PATH))
    logger.debug("* Project destination directory: {}".format(
                settings.TARGET_PATH))
    logger.debug("* Exclude patterns: {}".format(
                settings.EXCLUDES))

    # Watcher settings
    watcher_templates_patterns = {
        'patterns': ['*.scss'],
        'ignore_patterns': ['*.part'],
        'ignore_directories': False,
        'case_sensitive': True,
    }

    # Init inspector instance shared through all handlers
    inspector = ScssInspector()

    # Registering event handlers to observer
    observer = Observer()
    project_handler = WatchdogProjectEventHandler(settings, logger, inspector,
                                                  **watcher_templates_patterns)

    lib_handler = WatchdogLibraryEventHandler(settings, logger, inspector,
                                              **watcher_templates_patterns)

    # Observe source directory
    observer.schedule(project_handler, settings.SOURCES_PATH, recursive=True)

    # Also observe libraries directories
    for libpath in settings.LIBRARY_PATHS:
        observer.schedule(lib_handler, libpath, recursive=True)

    # Start watching
    logger.warning("Launching the watcher, use CTRL+C to stop it")
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        logger.warning("CTRL+C used, stopping..")
        observer.stop()

    observer.join()



# -*- coding: utf-8 -*-
import os
import pytest


def test_finder_conditions_001(settings, finder):
    """finder.ScssFinder: Filter files from partials"""
    sources = [
        '/home/foo/hip.scss',
        '/home/foo/_hop.scss',
        '/home/food/pizza.scss',
        '/home/food/_pasta.scss',
        '/home/food/drink/_azzuro.scss',
    ]

    results = [filepath for filepath in sources \
               if finder.match_conditions(filepath)]


    assert results == [
        '/home/foo/hip.scss',
        '/home/food/pizza.scss',
    ]


def test_finder_conditions_002(settings, finder):
    """finder.ScssFinder: Exclude libdirs case 1"""
    sources = [
        '/home/food/pizza.scss',
        '/home/bar/hop.scss',
        '/home/foo/hip.scss',
    ]

    results = []

    for filepath in sources:
        if not finder.match_conditions(filepath,
            sourcedir=None,
            nopartial=False,
            exclude_patterns=[],
            excluded_libdirs=["/home/food", "/home/bar", ]
        ):
            continue

        results.append(filepath)


    assert results == [
        '/home/foo/hip.scss',
    ]


def test_finder_conditions_003(settings, finder):
    """finder.ScssFinder: Exclude libdirs case 2"""
    sources = [
        '/home/food/pizza.scss',
        '/home/bar/hop.scss',
        '/home/foo/hip.scss',
    ]

    results = []

    for filepath in sources:
        if not finder.match_conditions(filepath,
            sourcedir=None,
            nopartial=False,
            exclude_patterns=[],
            excluded_libdirs=["/home/foo"]
        ):
            continue

        results.append(filepath)


    assert results == [
        '/home/food/pizza.scss',
        '/home/bar/hop.scss',
    ]


def test_finder_conditions_004(settings, finder):
    """finder.ScssFinder: Exclude from patterns but let pass partials"""
    sources = [
        '/home/foo/hip.scss',
        '/home/foo/_hop.scss',
        '/home/food/pizza.scss',
        '/home/food/_pasta.scss',
        '/home/food/drink/_azzuro.scss',
    ]

    results = []

    for filepath in sources:
        if not finder.match_conditions(filepath,
            sourcedir="/home",
            nopartial=False,
            exclude_patterns=[
                "food/*.scss",
            ],
            excluded_libdirs=[]
        ):
            continue

        results.append(filepath)


    assert results == [
        '/home/foo/hip.scss',
        '/home/foo/_hop.scss',
    ]


def test_finder_conditions_005(settings, finder):
    """finder.ScssFinder: Filter files from partials and libdirs"""
    sources = [
        '/home/foo/hip.scss',
        '/home/foo/_hop.scss',
        '/home/food/pizza.scss',
        '/home/food/_pasta.scss',
        '/home/food/drink/_azzuro.scss',
        '/etc/sound/bim.scss',
        '/etc/sound/_bam.scss',
        '/etc/sound/boum/kaboom.scss',
    ]

    results = []

    for filepath in sources:
        if not finder.match_conditions(filepath,
            sourcedir="/home/foo",
            exclude_patterns=[],
            excluded_libdirs=["/etc/sound", "/home/food"]
        ):
            continue

        results.append(filepath)


    assert results == [
        '/home/foo/hip.scss',
    ]



# -*- coding: utf-8 -*-
"""
Settings model
==============

This define the model object containing settings that will be passed to
interfaces.

"""
import copy

from boussole.conf import DEFAULT_SETTINGS


class Settings(object):
    """
    Settings model object

    Class init method fills object attributes from default settings
    (``DEFAULT_SETTINGS``) then update it with initial settings if given.

    Settings are available as object attributes, there is also a private
    ``_settings`` attribute containing a dict of all stored settings. You are
    strongly advised to never directly manipulate the ``_settings`` attribute.
    Instead, allways use the ``update()`` method.

    Note:
        If you intend to manually open and fill a Settings instance, remember
        to allways use absolute paths in your settings. Relative path will
        cause issues in resolving that lead to wrong compilations.

    Keyword Arguments:
        initial (dict): A dictionnary of settings for initial values.
    """
    def __init__(self, initial={}):
        self._settings = copy.deepcopy(DEFAULT_SETTINGS)
        if initial:
            initial = self.clean(initial)
            self._settings.update(initial)
        self.set_settings(self._settings)

    def clean(self, settings):
        """
        Filter given settings to keep only key names available in
        ``DEFAULT_SETTINGS``.

        Args:
            settings (dict): Loaded settings.

        Returns:
            dict: Settings object filtered.

        """
        return {k: v for k, v in settings.items() if k in DEFAULT_SETTINGS}

    def set_settings(self, settings):
        """
        Set every given settings as object attributes.

        Args:
            settings (dict): Dictionnary of settings.

        """
        for k, v in settings.items():
            setattr(self, k, v)

    def update(self, settings):
        """
        Update object attributes from given settings

        Args:
            settings (dict): Dictionnary of elements to update settings.

        Returns:
            dict: Dictionnary of all current saved settings.

        """
        settings = self.clean(settings)

        # Update internal dict
        self._settings.update(settings)

        # Push every setting items as class object attributes
        self.set_settings(settings)

        return self._settings



# -*- coding: utf-8 -*-
import os
import pytest


def test_finder_changeextension_001(settings, finder):
    """finder.ScssFinder: Changing filename extension"""
    assert finder.change_extension("foo.scss", 'css') == "foo.css"


def test_finder_changeextension_002(settings, finder):
    """finder.ScssFinder: Changing filename extension"""
    assert finder.change_extension("foo.backup.scss", 'css') == "foo.backup.css"


def test_finder_changeextension_003(settings, finder):
    """finder.ScssFinder: Changing filename extension"""
    assert finder.change_extension("bar/foo.scss", 'css') == "bar/foo.css"


def test_finder_changeextension_004(settings, finder):
    """finder.ScssFinder: Changing filename extension"""
    assert finder.change_extension("/home/bar/foo.scss", 'css') == "/home/bar/foo.css"


def test_finder_changeextension_005(settings, finder):
    """finder.ScssFinder: Changing filename extension"""
    assert finder.change_extension("/home/bar/foo.backup.scss", 'css') == "/home/bar/foo.backup.css"



"""
Logging
=======
"""
import logging


# Default logger conf
BOUSSOLE_LOGGER_CONF = (
    ('DEBUG', 'cyan'),
    ('INFO', 'green'),
    ('WARNING', 'yellow'),
    ('ERROR', 'red'),
)


def init_logger(level, printout=True):
    """
    Initialize app logger to configure its level/handler/formatter/etc..

    Todo:
        * Colors using "python-colorlog";
        * We need a "notice" level higher than ERROR (so it's allways
          displayed);
        * A mean to raise click.Abort when ERROR is used;

    Args:
        level (str): Level name (``debug``, ``info``, etc..).

    Keyword Arguments:
        printout (bool): If False, logs will never be outputed.

    Returns:
        logging.Logger: Application logger.
    """
    root_logger = logging.getLogger("boussole")
    root_logger.setLevel(level)

    if not printout:
        from StringIO import StringIO
        dummystream = StringIO()
        root_logger.addHandler(logging.StreamHandler(dummystream))
    else:
        root_logger.addHandler(logging.StreamHandler())

    return root_logger



# -*- coding: utf-8 -*-
import os
import copy
import pytest

from boussole.conf.base_backend import SettingsBackendBase


def test_conf_backend_base_clean_001(settings, sample_project_settings):
    """conf.base_backend.SettingsBackendBase: Ensure cleaning dont drop anything"""
    backend = SettingsBackendBase()

    assert backend.clean(sample_project_settings) == sample_project_settings



# -*- coding: utf-8 -*-
"""
Watch commandline is currently not tested because of some limits with
``CliRunner`` that is not able to communicate with invoked command, so we
can't send interrupt with "CTRL+C". Also it may low down tests running.
"""



# -*- coding: utf-8 -*-
import os
import pytest

from boussole.exceptions import UnresolvablePath
from boussole.exceptions import UnclearResolution

def test_resolver_resolve_basic(settings, parser, resolver):
    """resolver.ImportPathsResolver: Resolve paths from basic sample"""
    sourcepath = os.path.join(settings.sample_path, 'main_basic.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())

    resolved_paths = resolver.resolve(sourcepath, finded_paths)

    assert resolved_paths == [
        os.path.join(settings.sample_path, '_vendor.scss'),
        os.path.join(settings.sample_path, '_empty.scss'),
    ]


def test_resolver_resolve_library(settings, parser, resolver):
    """resolver.ImportPathsResolver: Resolve paths from main_using_libs.scss
    that use included libraries"""
    sourcepath = os.path.join(settings.sample_path, 'main_using_libs.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())

    resolved_paths = resolver.resolve(
        sourcepath,
        finded_paths,
        library_paths=settings.libraries_fixture_paths
    )

    assert resolved_paths == [
        os.path.join(settings.lib2_path, 'addons/_some_addon.scss'),
        os.path.join(settings.sample_path, 'main_basic.scss'),
        os.path.join(settings.sample_path, 'components/_webfont.scss'),
        os.path.join(settings.lib1_path, 'library_1_fullstack.scss'),
    ]


def test_resolver_resolve_commented(settings, parser, resolver):
    """resolver.ImportPathsResolver: Resolve paths from sample with comments"""
    sourcepath = os.path.join(settings.sample_path, 'main_commented.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())

    resolved_paths = resolver.resolve(sourcepath, finded_paths)

    assert resolved_paths == [
        os.path.join(settings.sample_path, '_vendor.scss'),
        os.path.join(settings.sample_path, 'components/_filename_test_1.scss'),
        os.path.join(settings.sample_path, '_empty.scss'),
        os.path.join(settings.sample_path, 'components/_webfont.scss'),
        os.path.join(settings.sample_path, 'components/_filename_test_2.scss'),
    ]


def test_resolver_resolve_error_unresolvable(settings, parser, resolver):
    """resolver.ImportPathsResolver: Exception on wrong import path"""
    sourcepath = os.path.join(settings.sample_path, 'main_error.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())

    with pytest.raises(UnresolvablePath):
        resolver.resolve(
            sourcepath,
            finded_paths,
            library_paths=settings.libraries_fixture_paths
        )


def test_resolver_resolve_error_unclear_001(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates on unclear resolution"""
    sourcepath = os.path.join(settings.sample_path, 'main_twins_1.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())

    with pytest.raises(UnclearResolution):
        resolver.resolve(
            sourcepath,
            finded_paths,
            library_paths=settings.libraries_fixture_paths
        )


def test_resolver_resolve_error_unclear_002(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates on unclear resolution"""
    sourcepath = os.path.join(settings.sample_path, 'main_twins_2.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())

    with pytest.raises(UnclearResolution):
        resolver.resolve(
            sourcepath,
            finded_paths,
            library_paths=settings.libraries_fixture_paths
        )


def test_resolver_resolve_error_unclear_003(settings, parser, resolver):
    """resolver.ImportPathsResolver: Check candidates on explicit resolution"""
    sourcepath = os.path.join(settings.sample_path, 'main_twins_3.scss')
    with open(sourcepath) as fp:
        finded_paths = parser.parse(fp.read())

    results = resolver.resolve(
        sourcepath,
        finded_paths,
        library_paths=settings.libraries_fixture_paths
    )

    assert results == [
        os.path.join(settings.sample_path, '_vendor.scss'),
        os.path.join(settings.sample_path, 'components/_twin_3.scss'),
    ]
